import sys
from typing import Tuple, List
from ngsolve import *
import netgen
from netgen.geom2d import unit_square

# Config
MPI: bool = False
if len(sys.argv) >= 2 and sys.argv[1] == "mpi":
	MPI: bool = True
T: int = 1 # end time
tau: float = 1e-1 # time step size

# Global variables
curTime: float = 0

# Little helper
def directedVectorField(
	x: CoefficientFunction, 
	y: CoefficientFunction, 
	radius: float = 0.1, 
	midpoint: Tuple[float, float] = (0.5, 0.5), 
	mainDirection: Tuple[float, float] = (0.0, 1.0)
):
	mainDirectionCF = CoefficientFunction(mainDirection)
	angleVar = 0.1
	distVar = 0.1

	direction = CoefficientFunction((x - midpoint[0], y - midpoint[1]))
	dist = direction.Norm()
	scProd = InnerProduct(direction, mainDirectionCF)
	preangle = scProd / dist
	angle = acos(IfPos(preangle-1, 1.0, IfPos(-1-preangle, -1.0, preangle)))

	return exp(-angle**2 / 2 / angleVar) * \
		exp(-(dist-radius)**2 / 2 / distVar) / dist * direction

def vtkOut(quantity: str, gfs: List[GridFunction], labels: List[str]):
	if MPI:
		vtk = VTKOutput(
			ma = mesh, 
			coefs = gfs, 
			names = labels,
			filename = "data/{:s}/{:.6f}_p{:d}".format(quantity, curTime, mpi_world.rank),
			subdivision = 2
		)
	else:
		vtk = VTKOutput(
			ma = mesh, 
			coefs = gfs, 
			names = labels,
			filename = "data/{:s}/{:.6f}".format(quantity, curTime), 
			subdivision = 2
		)
	vtk.Do()

# Distributing the mesh if necessary
if MPI:
	if mpi_world.rank == 0:
		ngmesh = Mesh("mesh.vol").ngmesh
		ngmesh.Distribute(mpi_world)
	else:
		ngmesh = netgen.meshing.Mesh.Receive(mpi_world)
		ngmesh.SetGeometry(unit_square)
	mesh = Mesh(ngmesh)
else:
	mesh: Mesh = Mesh("mesh.vol")

# Setting up variational formulation
V = VectorH1(mesh, order=2, dirichlet="top|bottom|left|right")
P = H1(mesh)
L = NumberSpace(mesh)

FES = FESpace([V, P, L])

v, p, lam = FES.TrialFunction()
w, q, mu = FES.TestFunction()

# Instationary Stokes equation
# pressure in enforced to have zero mean
stokesExpr = InnerProduct(grad(v), grad(w))
stokesExpr += -p * div(w)
stokesExpr += q * div(v)
stokesExpr += mu * p
stokesExpr += lam * q

massExpr = v * w

# Prepare for Euler
sol = GridFunction(FES)
sol.components[0].Set((0,0))
sol.components[1].Set(0)
sol.components[2].Set(0)
psol = GridFunction(FES)
psol.vec.data = sol.vec

rhs = LinearForm(FES)
rhs += SymbolicLFI(tau * directedVectorField(x, y) * w)

mass = BilinearForm(FES, nonassemble = True)
mass += SymbolicBFI(massExpr)

MStar = BilinearForm(FES)
MStar += SymbolicBFI(massExpr)
MStar += SymbolicBFI(tau * stokesExpr)

rhs.Assemble()
MStar.Assemble()

# Semi-Implicit Euler
with TaskManager():
	while curTime <= T:
		vtkOut("velocity", [sol.components[0]], ["v"])

		tmp = sol.vec.CreateVector()

		mass.Apply(sol.vec, tmp)
		tmp.data += rhs.vec
		sol.vec.data = MStar.mat.Inverse(FES.FreeDofs(), inverse = "pardiso") * tmp

		curTime += tau
